"""
Version : 1.0 (06-09-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.utils.newton import newton_


class LogBarrier:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is he logarithmic barrier defined as:

                /  - ( log(x-a) + log(b-x) )    if  a < x < b
         f(x)= |
               \  + INF                                      otherwise

            with a < b

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>LogBarriers()(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same shape
    as the input 'x'. So, the command >>>LogBarriers().prox(x)   will return
    an array with the same shape as 'x'

     INPUTS
    ========
     x     - scalar or ND array
     a     - scalar or ND array with the same size as 'x' (default: a=1)
     b     - scalar or ND array with the same size as 'x' (default: b=1)
     gamma - positive, scalar or ND array with the same size as 'x' (default: gamma=1)

    =======
    Examples
    ========

     Evaluate the 'direct' function :

     >>>  LogBarrier()( 0.5 )
      1.386294361119890

      Compute the result element-wise for vector inputs :

     >>> LogBarrier(a=[-2, 0, -1], b=[2, 3, 5])( [-1, 0.3, 0.25] )
     -2.6691794267132

     Compute the proximity operator at a given point :

     >>>  LogBarrier(a=[-2, 0, -1], b=[2, 3, 5]).prox(  [-2, 3, 4 ])
     array([-1.54269095,  2.37870272,  3.26376922])

     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function
     'gamma*f'

     >>>  LogBarrier(a=[-2, 0, -1], b=[2, 3, 5]).prox( [-2, 3, 4 ], gamma=2.5)
     array([-0.80311322,  1.93746944,  3.20331613])
    """

    def __init__(
            self,
            a: Union[float, np.ndarray] = 0,
            b: Union[float, np.ndarray] = 1
    ):
        if np.size(a) > 1 and (not isinstance(a, np.ndarray)):
            a = np.array(a)
        if np.size(b) > 1 and (not isinstance(b, np.ndarray)):
            b = np.array(b)
        if np.any(b < a):
            raise ValueError(
                "'b' (or all its components if it is an array) must be greater than 'a'"
            )
        self.a = a
        self.b = b

    def prox(self, x, gamma: Union[float, np.ndarray] = 1.0):
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        self._check(x, gamma)
        a = self.a
        b = self.b

        def polynom_phi(t):
            return (
                t**3
                - (a + b + x) * t**2
                + (a * b + (a + b) * x - 2 * gamma) * t
                - a * b * x
                + (a + b) * gamma
            )

        def der_polynom_phi(t):
            return 3 * t**2 - 2 * (a + b + x) * t + a * b + (a + b) * x - 2 * gamma

        # starting point
        prox_x = (a + b) / 2
        prox_x = newton_(polynom_phi, fp=der_polynom_phi, x0=prox_x, low=a, high=b)

        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        a = self.a
        b = self.b
        self._check(x, 1)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        result = np.zeros_like(x)
        mask = np.logical_and(x > a, x < b)
        aa = a
        bb = b
        if np.size(a) > 1:
            aa = a[mask]
        if np.size(b) > 1:
            bb = b[mask]

        result[mask] = -np.log(x[mask] - aa) - np.log(bb - x[mask])

        result[np.logical_not(mask)] = np.inf

        return np.sum(result)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its components if it is an array)" +
                " must be strictly positive"
            )

        if (np.size(gamma) > 1) and (np.shape(gamma) != np.shape(x)):
            raise ValueError(
                "''gamma' must be either scalar or have the same size as 'x'"
            )

        if (np.size(self.a) > 1) and (np.shape(self.a) != np.shape(x)):
            raise ValueError("'a' must be either scalar or the same size as 'x'")

        if (np.size(self.b) > 1) and (np.shape(self.b) != np.shape(x)):
            raise ValueError("'b' must be either scalar or the same size as 'x'")



"""
Version : 1.0 (27-04-2017), updated (08-06-2022).

Authors  : Giovanni Chierchia and Mbaye Diongue

Copyright (C) 2017

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class Bathtub:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is 'bathtub' (or w-sensitive loss) function defined as:

                f(x) = max{|x|-w,0}

     with w > 0
     'gamma' is the scale factor

    INPUTS
     ========
      x     - scalar or ND array
      w     - positive, scalar or ND array with the same size as 'x'
      gamma - positive, scalar or ND array with the same size as 'x'
              [default value: gamma=1]

     When the input 'x' is an array, the output is computed element-wise :

         -When calling the function (and not the proximity operator) the result
         is computed element-wise SUM. So the command >>>Bathtub()(x) will
         return a scalar even if x is a vector.

         - But for the proximity operator (method 'prox'), the output has the
         same shape as the input 'x'.
         So, the command >>>Bathtub().prox(x)   will return an array with the
         same shape as 'x'

      ========
      Examples
      ========

      Evaluate the 'direct' function :
      >>> Bathtub()(5)
      4
      >>> Bathtub(2, gamma=3)([-4., 2., 3., 5])
      18.0

      Compute the proximity operator :

      >>> Bathtub(2).prox( 3)
      array([2])
      >>> Bathtub(3).prox([ -3., 1., 6.])
      array([-3.,  1.,  5.])

      Use a gamma factor 'gamma'>0 to commute the proximity operator of gamma*f

      >>> Bathtub(1).prox([ -3., 1., 6.], gamma=2)
      array([-1.,  1.,  4.])

      When 'gamma' has the same size as 'x' the computation is done element-wise:
      >>> Bathtub(w=1).prox([ -3., 1., 6.], gamma=[1, 2, 3])
      array([-2.,  1.,  3.])
    """

    def __init__(self, w: Union[float, np.ndarray]):
        if np.size(w) > 1 and (not isinstance(w, np.ndarray)):
            w = np.array(w)
        if np.any(w <= 0):
            raise ValueError("'w' must be positive")
        if np.size(w) > 1 and (not isinstance(w, np.ndarray)):
            w = np.array(w)
        self.w = w

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        if not isinstance(x, np.ndarray):
            x = np.array(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)

        self._check(x, gamma)
        # preliminaries
        abs_x = np.abs(x)
        sign_x = np.sign(x)

        # 2nd branch
        prox_x = self.w * sign_x

        # 1st branch
        mask = abs_x < self.w
        prox_x[mask] = x[mask]

        # 3rd branch
        mask = abs_x > self.w + gamma
        p3 = x - gamma * sign_x
        prox_x[mask] = p3[mask]
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        self._check_w(x)
        return np.sum(np.maximum(0, np.abs(x) - self.w))

    def _check_w(self, x):
        if (not np.isscalar(self.w)) and (
            np.shape(self.w) != np.shape(x) and (len(self.w) > 1)
        ):
            raise ValueError(
                "'w' (or all its components if it is an array) must be either scalar "
                + "or the same size as 'x'"
            )

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its components if it is an array) must"
                + "be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")
        if (np.size(self.w) > 1) and (np.size(self.w) != np.size(x)):
            raise ValueError("'w' must be either scalar or the same size as 'x'")



"""
Version : 1.0 (06-18-2022).

DEPENDENCIES:
     'solver_cubic.py' in the folder 'utils'

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.utils.solver_cubic import solver_cubic


class Cauchy:
    r"""Compute the proximity operator and the evaluation of the gamma*f.

      where the function f is defined as:

               f(x) = log( delta + x^2)


    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise.

     INPUTS
    ========
    x     - ND array
    gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    delta -positive, scalar or ND array with the same size as 'x'[default: delata=1e-10]

    Note: When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>Cauchy(delta=0.01, gamma=1)(x)
    will return
    a scalar even if x is a vector:

    >>> Cauchy(0.01, 1)(np.array([-1., 2., 3.]))
    3.597076643791892

    But as expected, >>>Cauchy(delta=0.01, gamma=1).prox(x) will
    return a vector with the same size as x:

    >>> Cauchy(0.01, 1).prox(np.array([-1., 2., 3.]))
    array([-0.00498744,  0.01005025,  0.0152717 ])
    """

    def __init__(
        self,
        delta: Union[np.ndarray, float] = 1e-10
    ):
        if np.any(delta <= 0):
            raise Exception(
                "'delta' (or all of its elements) must be strictly positive"
            )
        self.delta = delta

    def prox(self, x: np.ndarray, gamma: Union[np.ndarray, float] = 1.0) -> np.ndarray:
        self._check(x, gamma)
        delta = self.delta

        # Solve the cubic equation
        prox_x, p2, p3 = solver_cubic(1, -x, delta + 2 * gamma, -x * delta)

        def f(t):
            return 0.5 * np.abs(x - t) ** 2 + gamma * np.log(delta + t**2)

        mask = np.logical_and(np.isreal(p2), f(np.real(p2)) < f(prox_x))
        prox_x[mask] = np.real(p2[mask])

        mask = np.logical_and(np.isreal(p2), f(np.real(p3)) < f(prox_x))
        prox_x[mask] = np.real(p3[mask])

        return np.reshape(prox_x, np.shape(x))

    def __call__(self, x: np.ndarray) -> float:
        return np.sum(np.log(self.delta + x**2))

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise Exception("'gamma' (or all of its elements)" +
                            " must be strictly positive")
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise Exception("'gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 (06-09-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class SquaredBathtub:
    r"""Computes the proximity operator and the evaluation of gamma*f.

    Where f is the 'squared w-insensitive loss' function defined as:

            f(x) =  (max{|x|-w,0} )^2   with w > 0

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function, the output is a scalar (sum of the
    element-wise results ) .

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'.

     INPUTS
    ========
     x     - scalar or ND array
     w     - positive, scalar or ND array with the same size as 'x'
     gamma - positive, scalar or ND array with the same size as 'x' (default: gamma=1)

    =======
    Examples
    ========

     Evaluate the function  f:

     >>> SquaredBathtub(3)( 2 )
     0.0

      Compute the result element-wise for vector inputs :

     >>> SquaredBathtub(w=2)( [-1, 3, -5] )
     14.5

     Compute the proximity operator at a given point :

     >>> SquaredBathtub(2).prox(  [-2, 3, 4 ] )
     array([-2. ,  2.5,  3. ])

     Use a scale factor 'gamma'>0 to compute the proximity operator of
     the function 'gamma*f'

     >>> SquaredBathtub( 2).prox( [-2, 3, 4, 6 ], gamma=[1,2,3,3])
     array([-2.        ,  2.33333333,  2.5       ,  3.        ]
    """

    def __init__(self, w: Union[float, np.ndarray]):
        if np.size(w) > 1 and (not isinstance(w, np.ndarray)):
            w = np.array(w)
        if np.any(w <= 0):
            raise ValueError(
                "'w' (or all of its components if it is an array)"
                + " must be strictly positive"
            )
        self.w = w

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        self._check(x, gamma)

        # preliminaries
        abs_x = np.abs(x)
        sign_x = np.sign(x)

        # 2nd branch
        prox_x = (x + gamma * self.w * sign_x) / (gamma + 1)

        # 1st branch
        mask = abs_x <= self.w
        prox_x[mask] = x[mask]
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        return np.sum(0.5 * (np.maximum(np.abs(x) - self.w, 0)) ** 2)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its components if it is an array) must "
                + "be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")
        if (np.size(self.w) > 1) and (np.size(self.w) != np.size(x)):
            raise ValueError("w' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-08-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class ELUacti:
    r"""Compute the proximity operator and the evaluation of f.

    Where f is the function defined as:

                          /  0                              if  x>=0
                   f(x)= |  (x+w)*ln( (x+w)/w) - x -x^2/2   Iif -w < x < 0
                         \  w - w^2 /2                      if  x=-w
                         \  +INF                            otherwise

    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>ELUacti(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same shape
    as the input 'x'.So, the command >>>ELUacti.prox(x)   will return an array
    with the same shape as 'x'.

     INPUTS
    ========
     x     - sclar ND array
     w     - w>=1, scalar or ND array with the same size as 'x'

    =======
    Examples
    ========

    Evaluate the function f:

     >>> ELUacti(w=4)( -3 )
     -2.886294361119891
     >>> ELUacti(2)( [1, -1.5, 3] )
     -0.318147180559945

     Compute the proximity operator at a given point :

     >>> ELUacti(2).prox( 3)
     array([3.])
     >>> ELUacti(2).prox(  [-2, 3, 4, np.e] )
     array([-1.72932943,  3.        ,  4.        ,  2.71828183]
    """

    def __init__(
            self,
            w: Union[float, np.ndarray]
    ):
        if np.size(w) > 1 and (not isinstance(w, np.ndarray)):
            w = np.array(w)
        self.gamma = 1
        if np.any(w < 1):
            raise Exception(
                "'w' (or all of its components if it is an array) must"
                + " be greater or equal than 1"
            )
        self.w = w

    def prox(self, x: np.ndarray) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        w = self.w
        prox_x = w * (np.exp(x) - 1)
        mask = x >= 0
        prox_x[mask] = x[mask]
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        w = self.w
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        result = np.zeros(np.shape(x))
        mask = np.logical_and(x > -w, x < 0)
        ww = w
        if np.size(w) > 1:
            ww = w[mask]

        result[mask] = (
            (x[mask] + ww) * np.log((x[mask] + ww) / ww) - x[mask] - x[mask] ** 2 / 2
        )

        mask2 = x == -w
        if np.size(w) > 1:
            ww = w[mask2]
        result[mask2] = ww - ww**2 / 2
        result[x < -w] = np.inf
        return np.sum(self.gamma * result)



"""
Compute here the lambert_W function.

Created on Wed Jun  8 2022

Author: Mbaye Diongue
"""
import numpy as np


def lambert_W(x, branch=0):
    r"""Compute here the lambert_W function.

    Parameters:

    x: float or np.ndarray
    branch: int
    """
    # CHose a starting point
    if branch != -1:
        w = np.ones_like(x)
    else:
        w = -2 * np.ones_like(x)

    # Haley's method
    stop = False
    tol = 1e-10
    max_iter = 100
    i = 0
    while (not stop) and (i < max_iter):
        w_old = w
        e = np.exp(w)
        f = w * e - x
        w = w - f / (e * (w + 1) - (w + 2) * f / (2 * w + 2))
        if np.linalg.norm(w - w_old) < np.linalg.norm(w) * tol:
            stop = True
        i += 1

    return w



"""
Compute the proximity operator of some indicator functions.

Author: Mbaye DIONGUE (Jun 2022)
"""

from .BoxConstraint import *
from .EpiAbsValue import *
from .EpiDistance import *
from .EpiExp import *
from .EpiHingeLoss import *
from .EpiL2Norm import *
from .EpiLinfNorm import *
from .EpiLog import *
from .EpiMax import *
from .EpiPower import *
from .EpiSquaredHingeLoss import *
from .EpiSupport import *
from .ExtendedSimplex import *
from .Hyperplane import *
from .Hyperslab import *
from .L1_2Ball import *
from .L1Ball import *
from .L2Ball import *
from .LinfBall import *
from .LogProjection import *
from .MonotoneCone import *
from .PositiveL2Ball import *
from .Simplex import *


__all__ = [
    "BoxConstraint",
    "EpiAbsValue",
    "EpiDistance",
    "EpiExp",
    "EpiHingeLoss",
    "EpiL2Norm",
    "EpiLinfNorm",
    "EpiLog",
    "EpiMax",
    "EpiPower",
    "EpiSquaredHingeLoss",
    "EpiSupport",
    "ExtendedSimplex",
    "Hyperplane",
    "Hyperslab",
    "L1_2Ball",
    "L1Ball",
    "L2Ball",
    "LinfBall",
    "LogProjection",
    "MonotoneCone",
    "PositiveL2Ball",
    "Simplex"
]



"""
Version : 1.0 (06-18-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Optional
import numpy as np


class L0L2Norm:
    r"""Compute the proximity operator and the evaluation of the gamma*f.

    Where the function f is defined as:

                                   / 0      if ||x||_2=0
               f(x)= ||x||_2^0 =  |
                                  \  1      otherwise


    'gamma' is the scale factor

    INPUTS
    ========
    x     - ND array
    gamma - positive, scalar or ND array compatible with the blocks of 'x'
            [default: gamma=1]
    axis   - None or int, axis of block-wise processing [default: axis=None]
            axis = None --> 'x' is processed as a single vector [DEFAULT] In this
            case 'gamma' must be a scalar.
            axis >=0   --> 'x' is processed block-wise along the specified axis
            (0 -> rows, 1-> columns ect. In this case, 'gamma' must be singleton
            along 'axis'.
    """

    def __init__(
            self,
            axis: Optional[int] = None
    ):
        self.axis = axis

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        self._check(x, gamma)
        axis = self.axis
        sz0 = np.shape(x)
        if axis is None:
            x = np.reshape(x, (-1))
        sz = np.shape(x)
        axis = self.axis
        sz = np.array(sz, dtype=int)
        sz[axis] = 1
        if np.size(gamma) > 1:
            gamma = np.reshape(gamma, sz)
        l2_x2 = np.sum(x**2, axis=axis).reshape(sz)
        mask = l2_x2 >= 2 * gamma
        prox_x = x * mask
        return np.reshape(prox_x, sz0)

    def __call__(self, x: np.ndarray) -> float:
        axis = self.axis
        if np.size(x) <= 1:
            x = np.array([x])
        if axis is None:
            x = np.reshape(x, (-1))
        p = np.linalg.norm(x, ord=2, axis=axis)
        return np.sum(np.array(p != 0))

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its elements) must be strictly positive"
            )
        if self.axis is None and np.size(gamma) > 1:
            raise ValueError(
                "An 'axis' must be specified when" + " 'gamma' is not a scalar"
            )
        if np.size(gamma) <= 1:
            return
        sz = np.shape(x)
        if len(sz) <= 1:
            self.axis = None
        if len(sz) <= 1:
            raise ValueError("'gamma' must be scalar when 'x' is one dimensional")
        if len(sz) > 1 and self.axis is not None:
            sz = np.array(sz, dtype=int)
            sz[self.axis] = 1
            if np.prod(sz) != np.size(gamma):
                raise ValueError(
                    "The dimension of 'gamma' is not "
                    + "compatible with the blocks of 'x'"
                )



"""
Version : 1.0 ( 06-14-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Optional
import numpy as np


class PositiveL2Ball:
    """Compute the projection and the indicator of the positive L2-ball.

    Recall: every vector X belonging to the positive L2-ball verifies:

                        ||x||_2 <= eta and x >= 0

     INPUTS
    ========
     x         - ND array
     eta       - positive scalar
     axis  - int or None, direction of block-wise processing
            When the input 'x' is an array, the computation can vary as follows:
            - axis = None --> 'x' is processed as a single vector [DEFAULT]
            - axis >= 0 --> 'x' is processed block-wise along the specified axis
              (axis=0 -> rows, axis=1 -> columns etc.).
    """

    def __init__(
            self,
            eta: float,
            axis: Optional[int] = None
    ):
        if np.size(eta) > 1 or eta <= 0:
            raise Exception("'eta' must be a positive scalar")
        self.eta = eta
        self.gamma = 1.0
        self.axis = axis

    def prox(self, x: np.ndarray) -> np.ndarray:
        xplus = np.maximum(x, 0)
        l2_x_plus = np.linalg.norm(xplus, ord=2, axis=self.axis)
        if l2_x_plus <= self.eta:
            return xplus
        return self.eta / l2_x_plus * xplus

    def __call__(self, x: np.ndarray) -> float:
        """
        Indicate if the input 'x' is in the constraint set or not
        Parameters
        ----------
        x : np.ndarray

        Returns
        -------
        0      if the input 'x' is in the set
        +inf   otherwise
        """
        l2_x = np.linalg.norm(x, ord=2, axis=self.axis, keepdims=True)
        if np.all(l2_x <= self.eta) and np.all(x >= 0):
            return 0
        return np.inf



"""
Version : 1.0 ( 06-22-2022).

DEPENDENCIES:
    -'newton.py'  - located in the folder 'utils'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Tuple
import numpy as np
from proxop.utils.newton import newton_


class Ialpha:
    r"""Compute the proximity operator and the evaluation of gamma*D.

    Where D is the function defined as:


                  /  - ( x* y^{alpha-1})^{1/alpha}   if x >= 0 and y >= 0
        D(x,y) = |
                 \   + inf                             otherwise

   'gamma' is the scale factor

    When the inputs are arrays, the outputs are computed element-wise

    ========
     INPUTS
    ========
    x       - scalar or ND array
    y       - ND array with the same size as 'x' otherwise
    alpha   - >1, scalar or ND array [ default: alpha=2]
    gamma   - positive, scalar or ND array with the same size as 'x' [default: gamma=1]
    """

    def __init__(
            self,
            alpha: Union[float, np.ndarray] = 2
    ):
        if np.any(alpha <= 1) or np.size(alpha) > 1:
            raise ValueError("'alpha'  must be a scalar greater than 1")
        self.alpha = alpha

    def prox(
            self,
            x: np.ndarray,
            y: np.ndarray,
            gamma: Union[float, np.ndarray] = 1
    ) -> Tuple[np.ndarray, np.ndarray]:
        scale = gamma
        self._check(y, gamma)
        alpha = self.alpha
        if np.size(x) != np.size(y):
            raise Exception("'x' and 'y' must have the same size")

        # scalar-like inputs handling
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
            y = np.reshape(y, (-1))

        # 2nd branch
        sz = np.shape(x)
        prox_p = np.zeros(sz)
        prox_q = np.zeros(sz)

        # branch selection
        mask = np.logical_or(
            alpha * x >= scale,
            1 - alpha * y / (scale * (alpha - 1))
            < scale / (scale - alpha * x) ** (1 / (alpha - 1)),
        )
        xx = x[mask]
        yy = y[mask]

        if np.size(scale) > 1:
            scale = scale[mask]
        # newton's method
        a = alpha * xx / scale - 1
        b = alpha - 1 - alpha * yy / scale

        def fun_phi(t):
            return t ** (2 * alpha) + a * t ** (alpha + 1) + b * t - alpha + 1

        def der_fun_phi(t):
            return 2 * alpha * t ** (2 * alpha - 1) + alpha * a * t**alpha + b

        # root finding
        eps = 1e-10
        low = np.maximum(0, 1 - alpha * xx / scale) ** (1 / (alpha - 1)) + eps
        root = 50 / alpha + low
        root = newton_(fun_phi, fp=der_fun_phi, x0=root, low=low, high=np.inf)

        # special cases
        root[(a == 0) * (b == 0)] = 1

        # 1st branch
        prox_p[mask] = xx + scale / alpha * (root ** (alpha - 1) - 1)
        prox_q[mask] = yy + scale * (1 - 1 / alpha) * (1 / root - 1)
        return tuple([prox_p, prox_q])

    def __call__(self, x: np.ndarray, y: np.ndarray) -> float:
        if np.any(x < 0) or np.any(y < 0):
            return np.inf
        return np.sum((-(x**self.alpha) * y ** (1 - 1 / self.alpha)))

    def _check(self, x, gamma=1):
        if np.any(gamma <= 0):
            raise Exception("'gamma'  must be strictly positive")
        if np.size(gamma) > 1 and np.size(gamma) != np.size(x):
            ValueError(
                "'gamma' must be positive scalars or positive ND arrays" +
                " with the same size as 'x'"
            )



"""
Version : 1.0 ( 06-13-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Optional
import numpy as np


class Max:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the function defined as:

                        f(x) = max( x1, ..., xn)

    'gamma' is the scale factor

     INPUTS
    ========
     x         - scalar or ND array
     gamma     - positive, scalar or ND array compatible with the blocks of 'x'
                 [default: gamma=1]
     axis      - None or int, axis of block-wise processing [default: axis=None]
                  axis = None --> 'x' is processed as a single vector [DEFAULT] In this
                  case 'gamma' must be a scalar.
                  axis >=0   --> 'x' is processed block-wise along the specified axis
                  (0 -> rows, 1-> columns ect. In this case, 'gamma' must be singleton
                  along 'axis'.
    """

    def __init__(
            self,
            axis: Optional[int] = None
    ):
        if (axis is not None) and (axis < 0):
            axis = None
        self.axis = axis

    def prox(
            self,
            x: np.ndarray,
            gamma: Union[float, np.ndarray] = 1.0
    ) -> np.ndarray:
        scale = gamma
        self._check(x, gamma)
        axis = self.axis
        sz = np.shape(x)

        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        sz0 = np.shape(x)
        sz0 = np.array(sz0, dtype=int)
        sz0[axis] = 1
        if np.size(scale) > 1:
            scale = np.reshape(scale, sz0)

        # 0. Linearize
        if axis is None:
            x = np.reshape(x, (-1))

        # 1. Order the column elements (sort in decreasing order)
        sort_x = -np.sort(-x, axis=axis)

        # 2. Compute the partial sums: c(j) = ( sum(s(1:j)) - gamma ) / j
        ones_ = np.ones(np.shape(sort_x))

        cum_sum = (np.cumsum(sort_x, axis=axis) - scale) / np.cumsum(ones_, axis=axis)

        # 3. Find the index: n = max{ j \in {1,...,B} : s(j) > c(j) }
        mask = sort_x > cum_sum
        mat = (np.arange(np.size(mask))).reshape(np.shape(mask))
        mask = mask * mat
        ind_max = np.argmax(mask, axis=axis)

        if np.size(ind_max) <= 1:
            prox_x = np.reshape(cum_sum[ind_max], np.shape(scale))
        else:
            prox_x = np.minimum(np.diag(np.take(cum_sum, ind_max,
                                                axis=axis)).reshape(sz0), x)
            prox_x = np.reshape(prox_x, np.shape(x))

        # 4. Compute the prox
        prox_x = np.minimum(x, prox_x)

        # 5. Output is zero if 's(j) > c(j)' for all j
        prox_x = prox_x * ((1 - np.all(mask, axis)).reshape(sz0))

        # 6. revert back
        prox_x = np.reshape(prox_x, sz)

        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        return np.sum(np.max(x, axis=self.axis))

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its components if it is an array)" +
                " must be strictly positive"
            )
        if self.axis is None and np.size(gamma) > 1:
            raise ValueError(
                "'gamma' must be a scalar when the parameter 'axis' is equal to 'None'"
            )
        if np.size(gamma) <= 1:
            return
        sz = np.shape(x)
        if len(sz) <= 1:
            self.axis = None
        if len(sz) <= 1:
            raise ValueError(
                "'gamma' must be scalar when 'x' is one dimensional"
            )
        if len(sz) > 1 and (self.axis is not None):
            sz = np.array(sz, dtype=int)
            sz[self.axis] = 1
            if np.size(gamma) > 1 and (np.prod(sz) != np.size(gamma)):
                raise ValueError(
                    "The dimension of 'gamma' is not compatible" +
                    " with the blocks of 'x'"
                )



"""
Version : 1.0 ( 06-22-2022).

DEPENDENCIES:
     - 'PersSquare.py' located in the folder 'multi'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Tuple
import numpy as np
from proxop.multi.PersSquare import PersSquare


class PersHuber:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the function defined as:


                 /   rho*|y| - 0.5*xi*rho^2     if xi > 0 and |y| > xi*rho
      f(y,xi) = |    | x |^2/(2*xi)             if xi > 0 and |y| <= xi*rho
                |    rho * |y|                  if y=0
                \    + inf                      otherwise



    'gamma' is the scale factor

            When the inputs are arrays, the outputs are computed element-wise
    INPUTS
    ========
    y         - scalar or ND array
    xi        - scalar or ND array compatible with the blocks of 'y'
    rho        - positive, scalar or ND array with the same size as 'y'
    gamma     - positive, scalar or ND array compatible with the blocks of 'x'
                 [default: gamma=1]
    """

    def __init__(
            self,
            rho: Union[float, np.ndarray],
            xi: Union[float, np.ndarray]
    ):

        if np.any(rho <= 0):
            raise Exception("'rho'  must be strictly positive")
        if np.any(np.size(rho) != np.size(xi)):
            raise Exception("'rho'  and 'xi' must have the same size")
        if np.size(xi) <= 1:
            xi = np.array([xi])
        if np.size(rho) <= 1:
            rho = np.reshape(rho, (-1))
        self.xi = xi
        self.rho = rho

    def prox(
            self,
            y: np.ndarray,
            gamma: Union[float, np.ndarray]
    ) -> Tuple[np.ndarray, np.ndarray]:

        self._check(y, gamma)
        rho = self.rho
        xi = self.xi

        if np.size(gamma) <= 1:
            gamma = np.reshape(gamma, (-1))
        if np.size(y) <= 1:
            y = np.reshape(y, (-1))

        sz0 = np.shape(y)
        sz = np.shape(y)
        if np.size(y) > 1:
            if np.size(xi) <= 1:
                xi = xi * np.ones(sz)
            if np.size(rho) <= 1:
                rho = rho * np.ones(sz)
            if np.size(gamma) <= 1:
                gamma = xi * np.ones(sz)
        abs_y = np.abs(y)
        sign_y = np.sign(y)
        rho2 = rho**2

        # 4th branch
        y_ = y
        if np.size(xi) > 1:
            y_ = np.reshape(y, np.shape(y) + tuple([1]))
            (prox_p, prox_q) = PersSquare(xi, axis=len(y_.shape) - 1).prox(y_)
        else:
            (prox_p, prox_q) = PersSquare(xi[0]).prox(y_)
        prox_p = np.reshape(prox_p, sz0)
        prox_q = np.reshape(prox_q, np.shape(xi))

        # 1st branch
        mask = (2 * xi * gamma + y**2 <= 0) * (abs_y < gamma * rho)
        prox_p[mask] = 0
        prox_q[mask] = 0

        # 2nd branch
        mask = (xi <= -0.5 * gamma * rho2) * (abs_y > gamma * rho)
        scale_m = gamma
        rho_m = rho
        if np.size(gamma) > 1:
            scale_m = gamma[mask]
        if np.size(rho) > 1:
            rho_m = rho[mask]
        prox_p[mask] = y[mask] - rho_m * scale_m * sign_y[mask]

        # 3rd branch
        mask = (xi > -0.5 * gamma * rho2) * (
            abs_y > rho * xi + gamma * rho * (1 + 0.5 * rho2)
        )
        scale_m = gamma
        rho_m = rho
        xi_m = xi
        if np.size(gamma) > 1:
            scale_m = gamma[mask]
        if np.size(rho) > 1:
            rho_m = rho[mask]
        if np.size(xi) > 1:
            xi_m = xi[mask]
        prox_p[mask] = y[mask] - rho_m * scale_m * sign_y[mask]
        prox_q[mask] = xi_m + 0.5 * scale_m * rho_m**2

        return tuple([np.reshape(prox_p, np.shape(y)), prox_q])

    def __call__(self, y: np.ndarray) -> float:
        if np.any(y < 0):
            return np.inf

        if np.size(y) <= 1:
            y = np.reshape(y, (-1))
        xi = self.xi
        rho = self.rho

        res = rho * np.abs(y) - 0.5 * xi * rho**2
        mask = np.logical_and(xi > 0, np.abs(y) <= xi * rho)
        xi_m = xi
        if np.size(xi) > 1:
            xi_m = xi[mask]
        res[mask] = 0.5 * np.abs(y[mask]) ** 2 / xi_m

        mask = y == 0
        rho_m = rho
        if np.size(rho) > 1:
            rho_m = rho[mask]
        res[mask] = rho_m * y[mask]

        return np.sum(res)

    def _check(self, y, gamma):
        if np.any(gamma <= 0):
            raise Exception("'gamma'  must be strictly positive")
        if np.size(self.xi) > 1 and np.size(self.xi) != np.size(y):
            raise Exception(
                "'xi' must be scalar or an ND array with the same size as 'y'"
            )
        if np.size(self.rho) > 1 and np.size(self.rho) != np.size(y):
            raise Exception(
                "'rho' must be scalar or an ND array with the same size as 'y'"
            )



"""
Version : 1.0 (06-14-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""
import numpy as np


def prox_svd(x, gamma, prox_phi, hermitian=False):
    r"""Compute the proximity operator of a matrix function.

                        f(X) = gamma * prox_phi(s)

    Where X = U*diag(s)*V.T \in R^{M*N}  is the Singular Value decomposition of X

      INPUTS
     ========
    x       - ND array
    gamma   - positive, scalar or ND array compatible with the size of 'x'
    prox_phi - function handle with two arguments at least
    """
    # spectral decomposition
    u, s, vh = np.linalg.svd(x, full_matrices=False, hermitian=hermitian)
    # prox computation
    g = np.reshape(prox_phi(s, gamma), np.shape(s))
    sz = np.shape(x)
    if len(sz) < 3:  # spectral reconstruction in the 2D case
        return np.dot(u * g, vh)
    else:  # spectral reconstruction in the 4D case
        # szdim4= min( np.shape(u)[3], np.shape(vh)[3])
        return np.matmul(u, g[..., None] * vh)



"""
Version : 1.0 ( 06-23-2022).

DEPENDENCIES:
     -'lambert_W.py' in the folder 'utils'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.utils.lambert_W import lambert_W


class BregLpNorm:
    r"""Compute the proximity operator prox_{gamma*f}^{phi}.

    Where the functions f and phi are defined as follows:

               f(x)= (1/p)*|x|**p


                         /  u*log(u)-u     if u>0
              phi(u) =  |
                        \   +inf           otherwise

    Note: The function phi is a Legendre type function and induces a Bregman distance

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise :

    - When calling the functions f or phi (method 'phi') the result is computed
    element-wise SUM. So the command >>>BregLpNorm()(x) or >>>BregLpNorm().phi(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the
    same shape as the input 'x'. So, the command >>>BregLpNorm.prox(x)
    will return an array with the same shape as 'x'


    INPUTS
    ========
    x      - scalar or ND array
    p      - scalar >=1 [default: p=1]
    gamma  - positive, scalar or ND array with the same size as 'x' [default: gamma=1]

    =======
    Examples
    ========

    Evaluate the function f:

     >>> BregLpNorm(p=2)( [1, 2, 3] )
     7.0
     >>> BregLpNorm()( [-2, 3, 4]) # default p=1
     9.0

     Evaluate The function  phi

     >> BregLpNorm().phi( [2, 3, 4] )
     1.227308671603782

     Compute the proximity operator at a given point :

     >>> BregLpNorm().prox(  [-2, 3, 4, np.e] )
     array([ 0.04978707,  7.3890561 , 20.08553692,  5.57494152])
     >>> BregLpNorm(2).prox(  [-2, 3, 4, np.e] )
     array([0.12002824, 2.20794003, 2.92627106, 2.01677976])
    """

    def __init__(self, p: float = 1):
        if np.any(p < 1) or np.size(p) > 1:
            raise ValueError("'p'  must be a scalar grater or equal to 1")
        self.p = p

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        self._check(x, gamma)
        p = self.p

        if p == 1:
            return np.exp(x - gamma)

        lambert_x = np.zeros(np.shape(x))

        u = np.log(gamma * (p - 1)) + (p - 1) * x
        # approximation of the lambert_W function when the input is "too big"
        mask = u > 100
        lambert_x[mask] = u[mask] - u[mask] / (1 + u[mask]) * np.log(u[mask])
        # use directly the Lambert_W function otherwise
        mask = np.logical_not(mask)
        lambert_x[mask] = lambert_W(np.exp(u[mask]))

        return (lambert_x / (gamma * (p - 1))) ** (1 / (p - 1))

    def __call__(self, x: np.ndarray) -> float:
        return np.sum((1 / self.p) * (np.abs(x)) ** self.p)

    def phi(self, u: np.ndarray) -> float:
        if np.size(u) > 1 and (not isinstance(u, np.ndarray)):
            u = np.array(u)
        if np.size(u) <= 1:
            u = np.reshape(u, (-1))
        res = np.zeros(np.shape(u))
        mask = u > 0
        res[mask] = u[mask] * np.log(u[mask]) - u[mask]
        res[u <= 0] = np.inf
        return np.sum(res)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError("'gamma'  must be strictly positive")
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-13-2022).

DEPENDENCIES:
    -'newton.py'  - located in the folder 'utils'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union, Optional
import numpy as np
from proxop.utils.newton import newton_


class HyperslabBarrier:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the  indicator of a hyper-slab barrier function defined as:

                / -log( high - a.T*x)-log( a.T - low)   if low < u.T*x < high
         f(x) =|
               \   + inf                                otherwise


    'gamma' is the scale factor

     INPUTS
    ========
     x      - ND array
     a      - ND array with the same size as 'x'
     low    - scalar or ND array compatible with the blocks of 'x'[default: low  = -1]
     high   - scalar or ND array compatible with the blocks of 'x' [default: high = 1]
     gamma  - positive, scalar or ND array compatible with the blocks of 'x'
             [default: gamma = 1]
     axis   - None or int, axis of block-wise processing [default: axis=None]
                  axis = None --> 'x' is processed as a single vector [DEFAULT] In this
                  case 'gamma' must be a scalar.
                  axis >=0   --> 'x' is processed block-wise along the specified axis
                  (0 -> rows, 1-> columns ect. in this case, 'gamma' and 'epsilon'
                  must be singletons along 'axis')
    """

    def __init__(
            self,
            a: Union[float, np.ndarray],
            low: Union[float, np.ndarray] = -1,
            high: Union[float, np.ndarray] = 1,
            axis: Optional[int] = None
    ):
        if np.any(low > high):
            raise ValueError(" 'high' must be greater than 'low'")
        if np.size(low) <= 1:
            low = np.reshape(low, (-1))
            high = np.reshape(high, (-1))
        self.a = a
        self.low = low
        self.high = high
        self.axis = axis

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1) -> np.ndarray:
        self._check(x)
        scale = gamma
        low = self.low
        high = self.high
        sz0 = np.shape(x)
        sz0 = np.array(sz0, dtype=int)
        sz0[self.axis] = 1
        if np.size(scale) > 1:
            scale = np.reshape(scale, sz0)

        sp = np.sum(self.a * x, self.axis)  # scalar product
        l2_a = np.sum(self.a ** 2, axis=self.axis)

        def polynom_phi(t):
            return (
                    t ** 3
                    - (low + high + sp) * t ** 2
                    + (low * high + sp * (low + high) - 2 * scale * l2_a) * t
                    - low * high * sp
                    + scale * (low + high) * l2_a
            )

        def derive_polynom_phi(t):
            return (
                    3 * t ** 2
                    - 2 * (low + high + sp) * t
                    + (low * high + sp * (low + high) - 2 * scale * l2_a)
            )

        # starting point
        root_init = (low + high) / 2

        # Find the root of the polynom with the Newton method
        root = newton_(
            polynom_phi, fp=derive_polynom_phi, x0=root_init, low=low, high=high
        )

        coef = np.reshape((root - sp) / l2_a, sz0)
        return x + coef * self.a

    def __call__(self, x: np.ndarray) -> float:
        self._check(x)
        scalar_prod = np.sum(self.a * x)
        if np.size(scalar_prod) <= 1:
            scalar_prod = np.reshape(scalar_prod, (-1))
        res = np.inf * np.ones_like(scalar_prod)
        mask = (scalar_prod < self.high) * (scalar_prod > self.low)
        res[mask] = -np.log(self.high[mask] - scalar_prod[mask]) - np.log(
            scalar_prod[mask] - self.low[mask]
        )

        return np.sum(res)

    def _check(self, x, gamma=1):
        if np.size(gamma) <= 1:
            return
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its components if it is an array)" +
                " must be strictly positive"
            )
        if self.axis is None and np.size(gamma) > 1:
            raise ValueError(
                "'gamma' must be a scalar when the parameter 'axis' is equal to 'None'"
            )
        sz = np.shape(x)
        if len(sz) <= 1:
            self.axis = None

        if len(sz) <= 1:
            raise ValueError(
                "'gamma' must be scalar when 'x' is one dimensional"
            )

        if len(sz) > 1 and (self.axis is not None):
            sz = np.array(sz, dtype=int)
            sz[self.axis] = 1
            if np.size(gamma) > 1 and (np.prod(sz) != np.size(gamma)):
                raise ValueError(
                    "The dimension of 'gamma' is not compatible" +
                    " with the blocks of 'x'"
                )



"""
Version : 1.0 ( 06-09-2022).

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np


class ArgsinhActi:
    r"""Compute the proximity operator and the evaluation of f.

    where f is the inverse hyperbolic sine activation function:

                    f(x) = cosh(x) -(x^2)/2

    When the input 'x' is an array, the output is computed element-wise :

    - When calling the function  the result is computed element-wise SUM.
    So, the command >>>Thresholder()(x) will return a scalar even if x is a
    vector.

    - But for the proximity operator (method 'prox'), the output has the
    same shape as the input 'x'.So, the command >>>Thresholder().prox(x)
    will return an array with the same shape as 'x'


     INPUTS
    ========
     x     - scalar or ND array

    ========
    Examples
    ========

    Evaluate the 'direct' function :
    >>> ArgsinhActi()(-0.5)
     1.0026259652063807
    >>>ArgsinhActi()([-2., -0.5, 0, 0.25, 3])
    10.332646751947351

    Compute the proximity operator at a given point :
    >>> ArgsinhActi().prox(3)
     1.8184464592320668s
    >>> ArgsinhActi().prox([ -3., 0.5, 6.])
    array([-1.81844646,  0.48121183,  2.49177985])
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray) -> np.ndarray:
        return np.arcsinh(x)

    def __call__(self, x: np.ndarray) -> float:
        if not isinstance(x, np.ndarray):
            x = np.array(x)
        return np.sum(np.cosh(x) - 0.5 * x**2)



"""
Version : 1.0 ( 06-10-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2019

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class Hyperslab:
    """Compute the projection onto the hyper-slab.

    Where the hyper-slab is the set defined as:

                         low  <= u.T * x <= high

     where u.T denotes the transpose of u

     INPUTS
    ========
     x    - ND array
     u    - ND array with the same size as 'x'
     low  - scalar or ND array compatible with the blocks of 'x' [default: low=0]
     high - scalar or ND array compatible with the blocks of 'x' [default: high=1]
    """

    def __init__(
            self,
            u: np.ndarray,
            low: Union[float, np.ndarray] = 0.0,
            high: Union[float, np.ndarray] = 1.0
    ):
        if np.any(low > high):
            raise ValueError("'low' must be lower than 'high'")
        self.low = low
        self.high = high
        self.u = u

    def prox(self, x: np.ndarray) -> np.ndarray:
        """Compute the proximity operator (i.e. projection of 'x' on the hyper-slab).

        Parameters
        ----------
        x : np.ndarray

        Returns
        -------
        TYPE : np.ndarray with the same size as x
        """
        self._check(x)
        scalar_prod = np.sum(x * self.u)
        u = self.u
        if np.any(scalar_prod < self.low):
            return x + ((self.low - scalar_prod) / np.sum(u**2)) * u
        if np.any(scalar_prod > self.high):
            return x + ((self.high - scalar_prod) / np.sum(u**2)) * u
        return x

    # indicator of the hyperslab
    def __call__(self, x: np.ndarray) -> float:
        """
        Indicate if the input 'x' is in the constraint set or not.

        Parameters
        ----------
        x : np.ndarray

        Returns
        -------
        0      if the input 'x' is in the set
        +inf   otherwise
        """
        self._check(x)
        scalar_prod = np.sum(x * self.u)
        tol = 1e-10
        if np.all(scalar_prod >= self.low - tol) and np.all(
            scalar_prod <= self.high + tol
        ):
            return 0
        return np.inf

    def _check(self, x: np.ndarray):
        if (np.shape(self.u)) != np.shape(x):
            raise ValueError("'u' must have the same size as 'x'")



"""
Version : 1.0 (06-08-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class Thresholder:
    r"""Computes the proximity operator and the evaluation of gamma*f.

    Where f is the Thresholder (or Support) function defined as:

                    /  a * x   if x < 0
            f(x) = |   0                        if x = 0         with a <= b
                   \  b * x    otherwise

     'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function, the output is a scalar (sum of the
    element-wise results ) .

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'.

     INPUTS
    ========
     x     - scalar or ND array
     a     - scalar or ND array with the same size as 'x' [default: a=-1]
     b     - scalar or ND array with the same size as 'x' [default: b=1]
     gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1.0]

     ========
     Examples
     ========

     Evaluate the 'direct' function f (i.e compute f(x)  ):

     >>> Thresholder(-2, 2)(3)
     6
     >>> Thresholder(-2, 2, gamma=2)([3, 4, -2])
     36

     Compute the proximity operator at a given point :

     >>> Thresholder(-2, 2).prox( 3)
     1
     >>> Thresholder(-1, 2).prox([ -3., 1., 6.])
     array([-2.,  0.,  4.])

     Use a scale factor 'gamma'>0 to commute the proximity operator of gamma*f

     >>> Thresholder(-1, 2).prox([ -3., 1., 6.], gamma=2)
      array([-1.,  0.,  2.])
    """

    def __init__(
        self,
        a: Union[float, np.ndarray] = -1,
        b: Union[float, np.ndarray] = 1
    ):
        if np.size(a) > 1 and (not isinstance(a, np.ndarray)):
            a = np.array(a)
        if np.size(b) > 1 and (not isinstance(b, np.ndarray)):
            b = np.array(b)
        if np.any(a > b):
            raise Exception("'b' must be greater than 'a'")
        self.a = a
        self.b = b

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        if not isinstance(x, np.ndarray):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        self._check_a_b(x, gamma)
        return np.minimum(0, x - self.a * gamma) + np.maximum(0, x - self.b * gamma)

    def __call__(self, x: np.ndarray) -> float:
        if not isinstance(x, np.ndarray):
            x = np.array(x)
        return np.sum((self.a * np.minimum(0, x) + self.b * np.maximum(0, x)))

    def _check_a_b(self, x, gamma):
        if np.any(gamma <= 0):
            raise Exception(
                "'gamma' (or all of its components if it is an array)"
                + " must be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.shape(gamma) != np.shape(x)):
            raise Exception("'gamma' must be either scalar or the same size as 'x'")
        if (np.size(self.a) > 1) and (np.shape(self.a) != np.shape(x)):
            raise Exception("'a' must be either scalar or the same size as 'x'")
        if (np.size(self.b) > 1) and (np.shape(self.b) != np.shape(x)):
            raise Exception("'b' must be either scalar or the same size as 'x'")



"""
Version : 1.0 ( 06-13-2022).

DEPENDENCIES:
     - 'prox_svd.py' located in the folder 'utils'
     - 'fun_svd.py' located in the folder 'utils'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

import numpy as np
from proxop.utils.prox_svd import prox_svd
from proxop.utils.fun_svd import fun_svd


class NuclearNormRidge:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the function defined as:


            f(x) = (1/2)*||X||_F^2 + mu*R1(C) =(1/2)||S||_2^2 + mu*||s||_1

    where X = U*diag(s)*V.T \in R^{M*N}  (Singular Value decomposition)

    'gamma' is the scale factor

     INPUTS
    ========
     x         -  (M,N) -array_like ( representing an M*N matrix )
     mu        - positive scalar [default: mu=1]
    gamma     - positive scalar  [default: gamma=1]
    """

    def __init__(
            self,
            mu: float = 1
    ):
        if np.any(mu <= 0) or np.size(mu) > 1:
            raise Exception("'mu'  must be a strictly positive scalar")
        self.mu = mu

    def prox(self, x: np.ndarray, gamma: float = 1) -> np.ndarray:
        mu = self.mu
        self._check(x, gamma)

        def prox_phi(s, gam_):
            a = -mu / (1 + gam_)
            b = mu / (1 + gam_)
            ss = s / (1 + gam_)
            phi_s = np.minimum(0, ss - a * gam_) + np.maximum(0, ss - b * gam_)
            return phi_s

        return prox_svd(x, gamma, prox_phi)

    def __call__(self, x: np.ndarray) -> float:

        def fun_phi(s):
            return 0.5 * np.sum(s ** 2) + self.mu * np.sum(np.abs(s))

        return fun_svd(x, 1, fun_phi)

    def _check(self, x, gamma):
        if np.any(gamma <= 0) or np.size(gamma) > 1:
            raise Exception("'gamma'  must be a strictly positive scalar")
        if len(np.shape(x)) != 2:
            raise ValueError(
                "'x' must be an (M,N) -array_like ( representing a M*N matrix )"
            )



"""
Version : 1.0 ( 06-13-2022).

DEPENDENCIES:
     - 'fun_svd.py' located in the folder 'utils'
     - 'prox_svd.py' located in the folder 'utils'
     - 'AbsValue.py' located in the folder 'utils'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.scalar.AbsValue import AbsValue
from proxop.utils.prox_svd import prox_svd


class NuclearNorm:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the function defined as:


                        f(x) = ||X||_N = ||s||_1

            where X = U*diag(s)*V.T \in R^{M*N}

     INPUTS
    ========
     x         - (M,N) -array_like ( representing an M*N matrix )
     gamma     - positive, scalar or ND array compatible with the size of 'x'
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1) -> np.ndarray:
        if np.any(gamma <= 0):
            raise Exception(
                "'gamma' (or all of its components if it is an array)" +
                " must be strictly positive"
            )
        return prox_svd(x, gamma, AbsValue().prox)

    def __call__(self, x: np.ndarray) -> float:
        return np.linalg.norm(x, ord='nuc')



"""
Version : 1.0 ( 06-09-2022).

DEPENDENCIES :
     lambert_W.py - located in the folder 'utils'

Author  : Mbaye Diongue

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.utils.lambert_W import lambert_W


class Entropy:
    r"""Compute the proximity operator and the evaluation of gamma*f.

    Where f is the entropy function defined as:


                      /  x*log(x)            if  x> 0
               f(x)= |   0                              if x =0
                     \  +INF                           otherwise


     When the input 'x' is an array, the output is computed element-wise.

     'gamma' is the scale factor

     When the input 'x' is an array, the output is computed element-wise :

    -When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>Entropy(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the same
    shape as the input 'x'. So, the command >>>Entropy.prox(x)   will return
    an array with the same shape as 'x'.

      INPUTS
     ========
     x     - scalar or ND array-like
     gamma - positive, scalar or ND array with the same size as 'x' [default: gamma=1]

    =======
    Examples
    ========

     Evaluate the 'direct' function :

     >>> Entropy()( np.e )
     2.718281828459045

     Element-wise sum when the input is a vector:
     >>> Entropy()( [1, 0, 3, np.e] )
      6.01411869446337

     Compute the proximity operator at a given point :

     >>> Entropy().prox(  [-2, 3, 4, np.e] )
     array([0.04747849, 1.5571456 , 2.20794003, 1.38940572])

     Use a scale factor 'gamma'>0 to compute the proximity operator of  the
     function 'gamma*f' :

     >>> Entropy().prox( [-2, 3, 4, np.e ], gamma=2 )
     array([0.12700814, 1.        , 1.37015388, 0.90903194])
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        self._check(x, gamma)
        inv_gamma = 1 / gamma
        return gamma * lambert_W(inv_gamma * np.exp(inv_gamma * x - 1))

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        result = np.zeros(np.shape(x))
        mask = x > 0
        result[mask] = x[mask] * np.log(x[mask])
        mask2 = x < 0
        result[mask2] = np.inf
        return np.sum(result)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its elements if it is an array)"
                + "must be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 (06-07-2022).

DEPENDENCIES:
     'newton.py' in the folder 'utils'

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np
from proxop.utils.newton import newton_


class Hyperbolic:
    r"""Compute the proximity operator and evaluate of gamma*f.

    Where f is the hyperbolic function defined as:

               f(x) = sqrt ( x^2 + delta^2)

    'gamma' is the scale factor

     When the input 'x' is an array, the output is computed element-wise :

     -When calling the function (and not the proximity operator) the result
     is computed element-wise SUM. So the command >>>Hyperbolic()(x) will
     return a scalar even if x is a vector.

     - But for the proximity operator (method 'prox'), the output has the
     same shape as the input 'x'.So, the command >>>Hyperbolic().prox(x)   will
     return an array with the same shape as 'x'

     INPUTS
     ========
     x      - scalar or array
     delta  - strictly positive, scalar or array with the same size as 'x'
     gamma  - positive, scalar or array with the same size as 'x' [ default: gamma=1]

     =======
     Examples
     ========

     Evaluate the 'direct' function :

     >>> Hyperbolic(delta=3)( 4 )
     5.0

     Compute the result as element-wise sum for vector inputs :

     >>> Hyperbolic(delta=2, gamma=3)( [-1, 3, 0.25] )
     23.571551070115248

     Compute the proximity operator at a given point :

     >>> Hyperbolic(3).prox(  [-2, 3, 4 ])
     array([-1.54269095,  2.37870272,  3.26376922])

     Use a scale factor 'gamma'>0 to compute the proximity operator of  the
     function 'gamma*f'

     >>> Hyperbolic(3, gamma=2.5).prox( [-2, 3, 4, np.e ] )
     array([-1.12333244,  1.74370688,  2.42744522,  1.5630981 ])
    """

    def __init__(self, delta: Union[float, np.ndarray]):
        if np.any(delta <= 0):
            raise ValueError(
                "'delta' (or all its components if it is an array)"
                + " must be strictly positive"
            )
        self.delta = delta

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        scale = gamma
        self._check(x, gamma)
        abs_x = np.abs(x)

        def fun_phi(p):
            return (
                p**4
                + (-2 * abs_x) * p**3
                + (x**2 - scale**2 + self.delta**2) * p**2
                + (-2 * abs_x * self.delta**2) * p
                + (self.delta**2) * (x**2)
            )

        def der_phi(p):
            return (
                4 * p**3
                - 6 * abs_x * p**2
                + 2 * (x**2 - scale**2 + self.delta**2) * p
                + (-2 * abs_x * self.delta**2)
            )

        # starting point
        root = abs_x / 2
        root = newton_(fun_phi, fp=der_phi, x0=root, low=0, high=abs_x)

        return root * np.sign(x)

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        return np.sum(np.sqrt(x**2 + self.delta**2))

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its elements if it is an array)"
                + "must be strictly positive"
            )
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")



"""
Version : 1.0 (06-09-2022).

Author  : Mbaye DIONGUE

Copyright (C) 2022

This file is part of the codes provided at http://proximity-operator.net

By downloading and/or using any of these files, you implicitly agree to
all the terms of the license CeCill-B (available online).
"""

from typing import Union
import numpy as np


class SquaredHinge:
    r"""Computes the proximity operator and the evaluation of gamma*f.

    Where f is the hinge Loss function defined as:

                      f(x) = max{1-x,0}^2

    'gamma' is the scale factor

    When the input 'x' is an array, the output is computed element-wise :

    -When calling the function (and not the proximity operator) the result
    is computed element-wise SUM. So the command >>>HingeLoss()(x) will
    return a scalar even if x is a vector.

    - But for the proximity operator (method 'prox'), the output has the
    same shape as the input 'x'. So, the command >>>HingeLoss().prox(x)
    will return an array with the same shape as 'x'

     INPUTS
    ========
     x     - ND array
     gamma  - positive, scalar or array with the same size as 'x' [default: gamma=1]

    =======
    Examples
    ========

     Evaluate the 'direct' function :

     >>> SquaredHinge()(2)
     0

    The output is computed as element-wise sum for vector inputs :

     >>> SquaredHinge()([2., 0.25, -0.5, 0.])
     3.8125

     Compute the proximity operator at a given point :

     >>> SquaredHinge().prox(  [-2, 3, 4 ])
     array([0.,  3.,  4.])

     Use a scale factor 'gamma'>0 to compute the proximity operator of  the function
     'gamma*f':

     >>> SquaredHinge().prox( [-2, 3, 4, np.e ] , gamma=2.5)
     array([0.5       , 3.        , 4.        , 2.71828183]
    """

    def __init__(self):
        pass

    def prox(self, x: np.ndarray, gamma: Union[float, np.ndarray] = 1.0) -> np.ndarray:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        if np.size(gamma) > 1 and (not isinstance(gamma, np.ndarray)):
            gamma = np.array(gamma)
        if np.size(x) <= 1:
            x = np.reshape(x, (-1))

        self._check(x, gamma)
        prox_x = (2*gamma + x)/(2*gamma + 1)
        mask = np.logical_and(x >= 1 - gamma, x <= 1)
        prox_x[mask] = 1
        mask = x > 1
        prox_x[mask] = x[mask]
        return prox_x

    def __call__(self, x: np.ndarray) -> float:
        if np.size(x) > 1 and (not isinstance(x, np.ndarray)):
            x = np.array(x)
        fun_x = np.maximum(1 - x, 0)**2
        return np.sum(fun_x)

    def _check(self, x, gamma):
        if np.any(gamma <= 0):
            raise ValueError(
                "'gamma' (or all of its elements if it is an array)"
                + "must be strictly positive"
            )
        if (np.size(gamma) > 1) and (np.size(gamma) != np.size(x)):
            raise ValueError("gamma' must be either scalar or the same size as 'x'")

